package com.skyversation.poiaddr.util.fileTools;

import com.skyversation.poiaddr.entity.FeatureDiy;
import com.skyversation.poiaddr.entity.FileDataDto;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.opencsv.CSVReader;
import com.opencsv.exceptions.CsvException;
import org.apache.poi.ss.usermodel.*;
import org.apache.poi.util.IOUtils;
import org.springframework.web.multipart.MultipartFile;

import javax.servlet.http.HttpServletResponse;
import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * 文件内容解析
 */
public class ReadFileData {
    //    文件内容解析
    public static List<FileDataDto> ReadMultipartFile(MultipartFile multipartFile) throws IOException {
        List<FileDataDto> fileDataDtoList = new ArrayList<>();
//      首先判断文件类型，走不同的解析方法
        String fileName = multipartFile.getOriginalFilename();
        assert fileName != null;
        String extension = fileName.substring(fileName.lastIndexOf(".") + 1);
        extension = extension.toLowerCase();
        // 创建临时文件
        File file = convert(multipartFile);
        if ("geojson".equals(extension) || "json".equals(extension)) {
            fileDataDtoList = readGeojson(file);
        } else if ("csv".equals(extension)) {
            fileDataDtoList = readCsv(file);
        } else if ("xlsx".equals(extension) || "xls".equals(extension)) {
            fileDataDtoList = readExcel(file);
        }
        return fileDataDtoList;
    }

    public static List<FileDataDto> readGeojson(File file) throws IOException {
        List<FileDataDto> resultList = new ArrayList<>();
        ObjectMapper mapper = new ObjectMapper();
        List<FeatureDiy> features = mapper.readValue(file, new TypeReference<List<FeatureDiy>>() {
        });
        for (FeatureDiy featureDiy : features) {
            FileDataDto fileDataDto = new FileDataDto();
            fileDataDto.setGeometry(featureDiy.getGeometry());
            fileDataDto.setProperties(featureDiy.getProperties());
            resultList.add(fileDataDto);
        }
        return resultList;
    }

    //    csv文件内容解析
    public static List<FileDataDto> readCsv(File file) {
        List<FileDataDto> resultList = new ArrayList<>();
        try {
            // 读取CSV文件
            CSVReader reader = new CSVReader(new FileReader(file));
            List<String[]> csvData = reader.readAll();
            reader.close();
            String[] headers = new String[]{};
            int index = 0;
            for (String[] row : csvData) {
                int rowSize = row.length;
                if (index == 0) {
                    headers = new String[rowSize];
                    for (int i = 0; i < rowSize; i++) {
                        headers[i] = row[i];
                    }
                } else {
                    FileDataDto fileDataDto = new FileDataDto();
                    Map<String, Object> colData = new HashMap<>();
                    for (int i = 0; i < rowSize; i++) {
                        colData.put(headers[i], row[i]);
                    }
                    fileDataDto.setProperties(colData);
                    resultList.add(fileDataDto);
                }
                index++;
            }
        } catch (IOException | CsvException e) {
            e.printStackTrace();
        }
        return resultList;
    }

    //    xlsx文件内容读取
    public static List<FileDataDto> readExcel(File file) throws IOException {
        List<FileDataDto> resultList = new ArrayList<>();
        FileInputStream fis = new FileInputStream(file);
        IOUtils.setByteArrayMaxOverride(400000000);
        // 创建工作簿对象，用于代表整个Excel文件
        Workbook workbook = WorkbookFactory.create(fis);
        // 这里我们默认读取第一个工作表，如果需要读取指定名称或者索引的工作表可以进行相应修改
        Sheet sheet = workbook.getSheetAt(0);
        // 获取表头行
        Row headerRow = sheet.getRow(0);
        int headerSize = headerRow.getLastCellNum();

        // 遍历数据行（从第二行开始，第一行是表头）
        for (int rowIndex = 1; rowIndex <= sheet.getLastRowNum(); rowIndex++) {
            FileDataDto fileDataDto = new FileDataDto();
            Row currentRow = sheet.getRow(rowIndex);
            Map<String, Object> rowMap = new HashMap<>();
            for (int cellIndex = 0; cellIndex < headerSize; cellIndex++) {
                Cell headerCell = headerRow.getCell(cellIndex);
                Cell currentCell = currentRow.getCell(cellIndex);
                if (currentCell != null) {
                    currentCell.setCellType(CellType.STRING);
                    rowMap.put(getCellValue(headerCell).toString(), getCellValue(currentCell).toString());
                } else {
                    rowMap.put(getCellValue(headerCell).toString(), null);
                }
            }
            fileDataDto.setProperties(rowMap);
            resultList.add(fileDataDto);
        }

        workbook.close();
        fis.close();
        return resultList;
    }

    //    xlsx行内数据解析
    private static Object getCellValue(Cell cell) {
        if (cell == null) {
            return null;
        }
        CellType cellType = cell.getCellType();
        switch (cellType) {
            case STRING:
                return cell.getStringCellValue();
            case NUMERIC:
                if (DateUtil.isCellDateFormatted(cell)) {
                    return cell.getDateCellValue();
                } else {
                    return cell.getNumericCellValue();
                }
            case BOOLEAN:
                return cell.getBooleanCellValue();
            case FORMULA:
                return cell.getCellFormula();
            default:
                return null;
        }
    }

    //    文件格式转换
    public static File convert(MultipartFile multipartFile) throws IOException {
        // 获取原始文件名
        String originalFilename = multipartFile.getOriginalFilename();
        // 获取文件后缀
        String fileExtension = originalFilename.substring(originalFilename.lastIndexOf("."));
        // 创建文件对象
        File convFile = File.createTempFile("temp", fileExtension);
        // 将 MultipartFile 转换为 Path
        Path path = Paths.get(convFile.getAbsolutePath());
        // 复制文件内容
        Files.copy(multipartFile.getInputStream(), path, StandardCopyOption.REPLACE_EXISTING);
        return convFile;
    }

    //    根据文件地址返回文件流
    public static void outputFile(String filePath, HttpServletResponse response) throws IOException {
        File file = new File(filePath);
        if (file.exists()) {
            // 设置响应头
            response.setContentType("application/octet-stream");
            response.setHeader("Content-Disposition", "attachment; filename=\"" + file.getName() + "\"");
            response.setContentLength((int) file.length());

            // 获取文件输入流
            try (FileInputStream fis = new FileInputStream(file);
                 OutputStream os = response.getOutputStream()) {
                byte[] buffer = new byte[4096];
                int bytesRead;
                // 从文件输入流读取数据并写入响应输出流
                while ((bytesRead = fis.read(buffer)) != -1) {
                    os.write(buffer, 0, bytesRead);
                }
                os.flush();
            } catch (IOException e) {
                e.printStackTrace();
            }
        } else {
            // 文件不存在，返回 404 错误
            response.sendError(HttpServletResponse.SC_NOT_FOUND, "文件未找到");
        }
    }

}
